from typing import Any, Self

from pydantic import BaseModel

from ormagic import DBField

from .clients.client import Client, client_context
from .field_utils import (
    is_many_to_many_field,
    is_primary_key_field,
    prepare_where_conditions,
)
from .table_manager import (
    create_table,
    get_foreign_key_model,
    get_intermediate_table_name,
    update_table,
)


class ObjectNotFound(Exception):
    pass


class DBModel(BaseModel):
    id: int | None = DBField(primary_key=True)

    @classmethod
    def __pydantic_init_subclass__(cls, **kwargs: Any) -> None:
        super().__pydantic_init_subclass__(**kwargs)
        for field_name, field_info in cls.model_fields.items():
            if is_primary_key_field(field_info) and field_name != "id":
                cls.model_fields.pop("id")
                break

    @classmethod
    def create_table(cls) -> None:
        """Create a table in the database for the model."""
        with client_context() as client:
            create_table(
                client,
                cls._get_table_name(),
                cls._get_primary_key_field_name(),
                cls.model_fields,
            )

    @classmethod
    def update_table(cls) -> None:
        """Update the table in the database based on the model definition."""
        with client_context() as client:
            update_table(
                client,
                cls._get_table_name(),
                cls._get_primary_key_field_name(),
                cls.model_fields,
            )

    @classmethod
    def drop_table(cls) -> None:
        """Remove the table from the database."""
        with client_context() as client:
            client.drop_table(cls._get_table_name())

    def save(self) -> Self:
        """Save object to the database."""
        with client_context() as client:
            return (
                self._update(client)
                if self.is_object_exists(client)
                else self._insert(client)
            )

    @classmethod
    def get(cls, *args, **kwargs) -> Self:
        """Get an object from the database based on the given keyword arguments."""
        with client_context() as client:
            return cls(**cls._fetchone_raw_data(client, *args, **kwargs))

    @classmethod
    def filter(cls, *args, **kwargs) -> list[Self]:
        """Get objects from the database based on the given keyword arguments."""
        with client_context() as client:
            return [
                cls(**data) for data in cls._fetchall_raw_data(client, *args, **kwargs)
            ]

    @classmethod
    def all(cls, *args, **kwargs) -> list[Self]:
        """Get all objects from the database."""
        with client_context() as client:
            return [
                cls(**data) for data in cls._fetchall_raw_data(client, *args, **kwargs)
            ]

    def delete(self) -> None:
        """Delete the object from the database."""
        with client_context() as client:
            client.delete_rows(
                self._get_table_name(),
                f"{self._get_primary_key_field_name()}={self.model_id}",
            )

    def _insert(self, client: Client) -> Self:
        prepared_data = self._prepare_data_to_insert()
        row_id = client.insert_row_and_get_id(
            self._get_table_name(),
            list(prepared_data.keys()),
            list(prepared_data.values()),
            self._get_primary_key_field_name(),
        )
        setattr(self, self._get_primary_key_field_name(), row_id)
        self._update_many_to_many_intermediate_table(client)
        return self

    def _update(self, client: Client) -> Self:
        data = self._prepare_data_to_insert()
        data.pop(self._get_primary_key_field_name())
        client.update_rows(
            self._get_table_name(),
            data,
            f"{self._get_primary_key_field_name()}={self.model_id}",
        )
        self._update_many_to_many_intermediate_table(client)
        return self

    def _update_many_to_many_intermediate_table(self, client: Client) -> None:
        related_objects = []
        for field_name, field_info in self.model_fields.items():
            if is_many_to_many_field(field_info.annotation):
                related_objects.extend(iter(getattr(self, field_name)))
        if not related_objects:
            return
        table_name = self._get_table_name()
        related_table_name = related_objects[0]._get_table_name()
        intermediate_table_name = get_intermediate_table_name(
            client, table_name, related_table_name
        )
        client.delete_rows(intermediate_table_name, f"{table_name}_id={self.model_id}")
        for related_object in related_objects:
            if not related_object.model_id:
                related_object = related_object.save()
            client.insert_row(
                intermediate_table_name,
                [f"{table_name}_id", f"{related_table_name}_id"],
                [str(self.model_id), str(related_object.model_id)],
            )

    def _prepare_data_to_insert(self) -> dict[str, Any]:
        prepared_data = {}
        model_dict = self.model_dump()
        for field_name, field_info in self.model_fields.items():
            field_value = model_dict.get(field_name)
            if foreign_model := get_foreign_key_model(field_info.annotation):
                if isinstance(field_value, list):
                    continue
                elif not field_value:
                    prepared_data[field_name] = None
                elif not field_value[self._get_primary_key_field_name()]:
                    foreign_model = foreign_model(**field_value).save()
                    prepared_data[field_name] = foreign_model.model_id
                    setattr(
                        getattr(self, field_name),
                        foreign_model._get_primary_key_field_name(),
                        foreign_model.model_id,
                    )
                else:
                    prepared_data[field_name] = field_value[
                        self._get_primary_key_field_name()
                    ]
            elif field_name == self._get_primary_key_field_name() and not field_value:
                continue
            else:
                prepared_data[field_name] = field_value
        return prepared_data

    def is_object_exists(self, client: Client) -> bool:
        if not self.model_id:
            return False
        return client.is_row_exists(
            self._get_table_name(),
            f"{self._get_primary_key_field_name()}={self.model_id}",
        )

    @classmethod
    def _prepare_order_by(
        cls, order_by: str | list[str] | tuple[str] | set[str]
    ) -> str:
        if isinstance(order_by, (list, tuple, set)):
            return ", ".join(cls._prepare_order_by(field) for field in order_by)
        return f"{order_by[1:]} DESC" if order_by.startswith("-") else order_by

    @classmethod
    def _prepare_query_to_fetch_raw_data(cls, *args, **kwargs) -> tuple[str, list]:
        sql = f"SELECT * FROM {cls._get_table_name()}"
        where_conditions, params = prepare_where_conditions(*args, **kwargs)
        if where_conditions:
            sql += f" WHERE {where_conditions}"
        if order_by := kwargs.get("order_by"):
            order_by = cls._prepare_order_by(order_by)
            sql += f" ORDER BY {order_by}"
        if limit := kwargs.get("limit"):
            sql += f" LIMIT {limit}"
        if offset := kwargs.get("offset"):
            sql += f" OFFSET {offset}"
        return sql, params

    @classmethod
    def _process_many_to_many_data(
        cls, client: Client, annotation: Any, object_id: int
    ) -> list[dict[str, Any]]:
        table_name = cls._get_table_name()
        related_model = getattr(annotation, "__args__")[0]
        related_table_name = related_model._get_table_name()
        intermediate_table_name = get_intermediate_table_name(
            client, table_name, related_table_name
        )
        rows = client.fetchall(
            f"SELECT {related_table_name}_id FROM {intermediate_table_name} WHERE {table_name}_id=?",
            [object_id],
        )
        return [
            related_model._fetchone_raw_data(
                client, is_recursive_call=True, model_id=row[0]
            )
            for row in rows
        ]

    @classmethod
    def _process_raw_data(
        cls, client: Client, data: tuple, is_recursive_call: bool = False
    ) -> dict[str, Any]:
        data_dict = dict(zip(cls.model_fields.keys(), data))
        for key, field_info in cls.model_fields.items():
            if is_many_to_many_field(field_info.annotation):
                if is_recursive_call:
                    continue
                data_dict[key] = cls._process_many_to_many_data(
                    client,
                    field_info.annotation,
                    data_dict[cls._get_primary_key_field_name()],
                )
            elif not data_dict[key]:
                continue
            elif foreign_model := get_foreign_key_model(field_info.annotation):
                data_dict[key] = foreign_model._fetchone_raw_data(
                    client, model_id=data_dict[key]
                )
        return data_dict

    @classmethod
    def _fetchone_raw_data(
        cls,
        client: Client,
        is_recursive_call: bool = False,
        model_id: int | None = None,
        *args,
        **kwargs,
    ) -> dict[str, Any]:
        if model_id:
            kwargs[cls._get_primary_key_field_name()] = model_id
        sql, params = cls._prepare_query_to_fetch_raw_data(*args, **kwargs)
        if data := client.fetchone(sql, params):
            return cls._process_raw_data(client, data, is_recursive_call)
        else:
            raise ObjectNotFound

    @classmethod
    def _fetchall_raw_data(
        cls, client: Client, *args, **kwargs
    ) -> list[dict[str, Any]]:
        query, params = cls._prepare_query_to_fetch_raw_data(*args, **kwargs)
        data_list = client.fetchall(query, params)
        return [cls._process_raw_data(client, data) for data in data_list]

    @classmethod
    def _get_table_name(cls) -> str:
        return f"{cls.__name__.lower()}s"

    @property
    def model_id(self) -> int | None:
        return getattr(self, self._get_primary_key_field_name())

    @classmethod
    def _get_primary_key_field_name(cls) -> str:
        return next(
            (
                field_name
                for field_name, field_info in cls.model_fields.items()
                if is_primary_key_field(field_info)
            ),
            "",
        )



from contextlib import contextmanager
from typing import Any, Generator, Protocol

from ormagic.settings import Settings


class DatabaseNotSupported(Exception):
    pass


class Client(Protocol):
    def create_connection(self) -> Any: ...
    def execute(self, sql: str, parameters: list | None = None) -> Any: ...
    def close(self) -> None: ...
    def rollback(self) -> None: ...
    def commit(self) -> None: ...
    def create_table(self, table_name: str, columns: list[str]) -> None: ...
    def is_table_exists(self, table_name: str) -> bool: ...
    def get_column_names(self, table_name: str) -> list[str]: ...
    def drop_column(self, table_name: str, column_name: str) -> None: ...

    def rename_column(
        self, table_name: str, old_column_name: str, new_column_name: str
    ) -> None: ...
    def add_column(self, table_name: str, column_name: str) -> None: ...
    def drop_table(self, table_name: str) -> None: ...

    def update_rows(
        self, table_name: str, fields: dict[str, str], where: str
    ) -> None: ...
    def delete_rows(self, table_name: str, where: str) -> None: ...

    def insert_row(
        self, table_name: str, columns: list[str], values: list[str]
    ) -> None: ...

    def insert_row_and_get_id(
        self, table_name: str, columns: list[str], values: list[str], id_column: str
    ) -> int: ...

    def is_row_exists(self, table_name: str, where: str) -> bool: ...
    def fetchone(self, sql: str, parameters: list[Any] | None = None) -> Any: ...
    def fetchall(self, sql: str, parameters: list[Any] | None = None) -> list[Any]: ...


def get_client() -> Client:
    settings = Settings()
    if settings.db_type == "sqlite":
        from .sqlite import SQLiteClient

        return SQLiteClient()
    elif settings.db_type == "postgresql":
        from .postgres import PostgresClient

        return PostgresClient()
    raise DatabaseNotSupported(f"{settings.db_type} is not supported")


@contextmanager
def client_context() -> Generator[Client, None, None]:
    from ormagic.transactions import transaction

    if transaction._is_transaction:
        yield transaction._client
    else:
        client = get_client()
        client.create_connection()
        try:
            yield client
        finally:
            client.close()



from .field_utils import prepare_where_conditions


class Q:
    def __init__(self, *args, **kwargs):
        self.conditions, self.params = prepare_where_conditions(*args, **kwargs)

        if len(self.params) > 1:
            self.conditions = f"({self.conditions})"

    def __or__(self, other: "Q") -> "Q":
        self.conditions = f"{self.conditions} OR {other.conditions}"
        self.params.extend(other.params)
        return self

    def __and__(self, other: "Q") -> "Q":
        self.conditions = f"{self.conditions} AND {other.conditions}"
        self.params.extend(other.params)
        return self

    def __invert__(self) -> "Q":
        self.conditions = f"NOT ({self.conditions})"
        return self



import os


class SettingsError(Exception):
    pass


class Settings:
    def __init__(self):
        self.database_url = os.getenv("ORMAGIC_DATABASE_URL", "sqlite://db.sqlite3")
        if "://" not in self.database_url:
            raise SettingsError("Database URL is not valid")
        self.db_type = self.database_url.split("://")[0]
        self.path = self.database_url.split("://")[1]
        self.journal_mode = os.getenv("ORMAGIC_JOURNAL_MODE", "WAL")



from sqlite3 import Connection, Cursor, connect
from typing import Any

from ormagic.settings import Settings


class SQLiteClient:
    def create_connection(self) -> Connection:
        settings = Settings()
        connection = connect(settings.path, isolation_level=None)
        connection.execute("PRAGMA foreign_keys = ON")
        connection.execute(f"PRAGMA journal_mode = {settings.journal_mode}")
        self.connection = connection
        self.cursor = connection.cursor()
        return connection

    def execute(self, sql: str, parameters: list | None = None) -> Cursor:
        return (
            self.cursor.execute(sql, parameters)
            if parameters
            else self.cursor.execute(sql)
        )

    def close(self) -> None:
        self.connection.close()

    def rollback(self) -> None:
        self.connection.rollback()

    def commit(self) -> None:
        self.connection.commit()

    def create_table(self, table_name: str, columns: list[str]) -> None:
        self.execute(f"CREATE TABLE IF NOT EXISTS {table_name} ({', '.join(columns)})")

    def is_table_exists(self, table_name: str) -> bool:
        cursor = self.execute(
            f"SELECT count(*) FROM sqlite_master WHERE type='table' AND name='{table_name}'"
        )
        return cursor.fetchone()[0] == 1

    def get_column_names(self, table_name: str) -> list[str]:
        cursor = self.execute(f"PRAGMA table_info({table_name})")
        return [column[1] for column in cursor.fetchall()]

    def drop_column(self, table_name: str, column_name: str) -> None:
        self.execute(f"ALTER TABLE {table_name} DROP COLUMN {column_name}")

    def rename_column(
        self, table_name: str, old_column_name: str, new_column_name: str
    ) -> None:
        self.execute(
            f"ALTER TABLE {table_name} RENAME COLUMN {old_column_name} TO {new_column_name}"
        )

    def add_column(self, table_name: str, column_name: str) -> None:
        self.execute(f"ALTER TABLE {table_name} ADD COLUMN {column_name}")

    def drop_table(self, table_name: str) -> None:
        self.execute(f"DROP TABLE {table_name}")

    def update_rows(self, table_name: str, fields: dict[str, Any], where: str) -> None:
        set_fields = ", ".join(
            f"{field} = '{value}'" if value else f"{field} = NULL"
            for field, value in fields.items()
        )
        self.execute(f"UPDATE {table_name} SET {set_fields} WHERE {where}")

    def delete_rows(self, table_name: str, where: str) -> None:
        self.execute(f"DELETE FROM {table_name} WHERE {where}")

    def insert_row(
        self, table_name: str, columns: list[str], values: list[str]
    ) -> None:
        self.execute(
            f"INSERT INTO {table_name} ({', '.join(columns)}) VALUES ({', '.join(['?' for _ in values])})",
            values,
        )

    def insert_row_and_get_id(
        self, table_name: str, columns: list[str], values: list[str], id_column: str
    ) -> int:
        cursor = self.execute(
            f"INSERT INTO {table_name} ({', '.join(columns)}) VALUES ({', '.join(['?' for _ in values])}) RETURNING {id_column}",
            values,
        )
        return cursor.fetchone()[0]

    def is_row_exists(self, table_name: str, where: str) -> bool:
        cursor = self.execute(f"SELECT count(*) FROM {table_name} WHERE {where}")
        return cursor.fetchone()[0] == 1

    def fetchone(self, sql: str, parameters: list[Any] | None = None) -> Any:
        cursor = self.execute(sql, parameters)
        return cursor.fetchone()

    def fetchall(self, sql: str, parameters: list[Any] | None = None) -> list[Any]:
        cursor = self.execute(sql, parameters)
        return cursor.fetchall()



from .clients.client import Client, get_client


class transaction:
    _is_transaction = False
    _client: Client

    @classmethod
    def __enter__(cls):
        cls._is_transaction = True
        cls._client = get_client()
        cls._client.create_connection()
        cls._client.execute("BEGIN")

    @classmethod
    def __exit__(cls, exc_type, exc_value, traceback):
        cls._is_transaction = False
        if exc_type:
            cls._client.rollback()
        else:
            cls._client.commit()
        cls._client.close()



from typing import TYPE_CHECKING, Any, Type, get_args

from pydantic.fields import FieldInfo
from pydantic_core import PydanticUndefined

from ormagic.field_utils import (
    get_on_delete_action,
    is_primary_key_field,
    is_unique_field,
    transform_field_annotation_to_sql_type,
)

from .clients.client import Client
from .field_utils import (
    is_many_to_many_field,
)

if TYPE_CHECKING:  # pragma: no cover
    from .models import DBModel


class UpdateTableError(Exception):
    pass


def create_table(
    client: Client,
    table_name: str,
    primary_key: str,
    model_fields: dict[str, FieldInfo],
):
    columns = []
    related_tables = []
    for field_name, field_info in model_fields.items():
        if is_many_to_many_field(field_info.annotation):
            related_tables.append(getattr(field_info.annotation, "__args__")[0])
            continue
        columns.append(prepare_column_definition(field_name, field_info))
    client.create_table(table_name, columns)
    for related_table in related_tables:
        _create_intermediate_table(client, table_name, primary_key, related_table)


def update_table(
    client: Client,
    table_name: str,
    primary_key: str,
    model_fields: dict[str, FieldInfo],
) -> None:
    if not client.is_table_exists(table_name):
        return create_table(client, table_name, primary_key, model_fields)
    if _some_fields_is_unique(model_fields):
        raise UpdateTableError("Cannot add unique field to existing table")
    existing_columns = client.get_column_names(table_name)
    new_columns = _get_model_field_names(model_fields)
    if existing_columns == new_columns:
        return
    elif len(existing_columns) > len(new_columns):
        _drop_columns_from_existing_table(
            client, table_name, existing_columns, new_columns
        )
    elif len(existing_columns) == len(new_columns):
        return _rename_columns_in_existing_table(
            client, table_name, existing_columns, new_columns
        )
    _add_new_columns_to_existing_table(
        client, table_name, model_fields, existing_columns
    )


def _some_fields_is_unique(model_fields: dict[str, FieldInfo]) -> bool:
    return any(is_unique_field(field_info) for field_info in model_fields.values())


def _create_intermediate_table(
    client: Client,
    table_name: str,
    primary_key: str,
    related_table: "DBModel",
) -> None:
    related_table_name = related_table._get_table_name()
    related_primary_key = related_table._get_primary_key_field_name()
    if not client.is_table_exists(related_table_name):
        return
    client.create_table(
        f"{table_name}_{related_table_name}",
        [
            "id INTEGER PRIMARY KEY",
            f"{table_name}_id INTEGER",
            f"{related_table_name}_id INTEGER",
            f"FOREIGN KEY ({table_name}_id) REFERENCES {table_name}({primary_key}) ON DELETE CASCADE ON UPDATE CASCADE",
            f"FOREIGN KEY ({related_table_name}_id) REFERENCES {related_table_name}({related_primary_key}) ON DELETE CASCADE ON UPDATE CASCADE",
        ],
    )


def get_intermediate_table_name(
    client: Client, table_name: str, related_table_name: str
) -> str:
    if client.is_table_exists(f"{table_name}_{related_table_name}"):
        return f"{table_name}_{related_table_name}"
    else:
        return f"{related_table_name}_{table_name}"


def prepare_column_definition(field_name: str, field_info: FieldInfo) -> str:
    field_type = transform_field_annotation_to_sql_type(field_info.annotation)
    column_definition = f"{field_name} {field_type}"
    if field_info.default not in (PydanticUndefined, None):
        column_definition += f" DEFAULT '{field_info.default}'"
    if field_info.is_required():
        column_definition += " NOT NULL"
    if is_unique_field(field_info):
        column_definition += " UNIQUE"
    if foreign_model := get_foreign_key_model(field_info.annotation):
        action = get_on_delete_action(field_info)
        column_definition += f", FOREIGN KEY ({field_name}) REFERENCES {foreign_model._get_table_name()}({foreign_model._get_primary_key_field_name()}) ON UPDATE {action} ON DELETE {action}"
    if is_primary_key_field(field_info):
        column_definition += " PRIMARY KEY"
    return column_definition


def get_foreign_key_model(field_annotation: Any) -> Type | None:
    from .models import DBModel

    types_tuple = get_args(field_annotation)
    if not types_tuple and field_annotation and issubclass(field_annotation, DBModel):
        return field_annotation
    if types_tuple and issubclass(types_tuple[0], DBModel):
        return types_tuple[0]


def _get_model_field_names(model_fields: dict[str, FieldInfo]) -> list[str]:
    return list(model_fields.keys())


def _rename_columns_in_existing_table(
    client: Client, table_name: str, old_columns: list[str], new_columns: list[str]
) -> None:
    for old_column_name, new_column_name in dict(zip(old_columns, new_columns)).items():
        if old_column_name != new_column_name:
            client.rename_column(table_name, old_column_name, new_column_name)


def _add_new_columns_to_existing_table(
    client: Client,
    table_name: str,
    model_fields: dict[str, FieldInfo],
    existing_columns: list[str],
) -> None:
    for field_name, field_info in model_fields.items():
        if field_name in existing_columns:
            continue
        column_definition = prepare_column_definition(field_name, field_info)
        client.add_column(table_name, column_definition)


def _drop_columns_from_existing_table(
    client: Client, table_name: str, existing_columns: list[str], new_columns: list[str]
) -> None:
    columns_to_drop = set(existing_columns) - set(new_columns)
    for column_name in columns_to_drop:
        client.drop_column(table_name, column_name)
